@@ -1,13 +1,13 @@
 /************************************************************************
 *                                                                       *
 *             Star Traders: A Game of Interstellar Trading              *
-*                Copyright (C) 1990-2020, John Zaitseff                 *
+*                Copyright (C) 1990-2024, John Zaitseff                 *
 *                                                                       *
 ************************************************************************/
 
 /*
   Author: John Zaitseff <J.Zaitseff@zap.org.au>
-  $Id: 88564ec81a9cf9dcf0ba72b60f6d3d4b5f149e94 $
+  $Id: 27cb4ed268bce9d27041a162a4cab329f4798a46 $
 
   Star Traders is a simple game of interstellar trading, where the object
   of the game is to create companies, buy and sell shares, borrow and
@@ -117,6 +117,19 @@ static void show_version (void) __attribute__((noreturn));
 static void show_usage (int status) __attribute__((noreturn));
 
 
+/*
+  Function:   init_program_prelim - Initialise program preliminaries
+  Parameters: argc                - Same as passed to main()
+              argv                - Same as passed to main()
+  Returns:    (nothing)
+
+  This function initialises the program name, locale and message
+  catalogs.  It must be called before command line arguments are
+  processed.
+*/
+static void init_program_prelim (int argc, char *argv[]);
+
+
 /*
   Function:   init_program - Initialise program-wide functions
   Parameters: (none)
@@ -146,18 +159,8 @@ static void end_program (void);
 
 int main (int argc, char *argv[])
 {
-    // Strip off leading pathname components from program name
-    init_program_name(argv[0]);
-
-    // Initialise the locale
-    if (setlocale(LC_ALL, "") == NULL) {
-	err_exit("could not set locale "
-		 "(check LANG, LC_ALL and LANGUAGE in environment)");
-    }
-
-    // Use correct message catalogs for the locale
-    bindtextdomain(PACKAGE, LOCALEDIR);
-    textdomain(PACKAGE);
+    // Initialise program name, locale and message catalogs
+    init_program_prelim(argc, argv);
 
     // Process command line arguments
     process_cmdline(argc, argv);
@@ -293,7 +296,7 @@ This program is free software that is distributed under the terms of the\n\
 GNU General Public License, version 3 or later.  You are welcome to\n\
 modify and/or distribute it under certain conditions.  This program has\n\
 NO WARRANTY, to the extent permitted by law; see the License for details.\n\
-"), program_name, PACKAGE_VERSION, "1990-2020");
+"), program_name, PACKAGE_VERSION, "1990-2024");
 
     exit(EXIT_SUCCESS);
 }
@@ -361,6 +364,19 @@ playing that game.  If GAME is not specified, start a new game.\n\n\
 // These functions are documented at the start of this file
 
 
+/***********************************************************************/
+// init_program_prelim: Initialise program preliminaries
+
+void init_program_prelim (int argc, char *argv[])
+{
+    // Strip off leading pathname components from program name
+    init_program_name(argv[0]);
+
+    // Initialise the locale and message catalogs
+    init_locale();
+}
+
+
 /***********************************************************************/
 // init_program: Initialise program-wide functions
 
@@ -370,7 +386,7 @@ void init_program (void)
     init_rand();
 
     // Initialise locale-specific variables
-    init_locale();
+    init_locale_vars();
 
     // Initialise the terminal display
     init_screen();
