@@ -1,13 +1,13 @@
 /************************************************************************
 *                                                                       *
 *             Star Traders: A Game of Interstellar Trading              *
-*                Copyright (C) 1990-2020, John Zaitseff                 *
+*                Copyright (C) 1990-2024, John Zaitseff                 *
 *                                                                       *
 ************************************************************************/
 
 /*
   Author: John Zaitseff <J.Zaitseff@zap.org.au>
-  $Id: 03bd86c1924cb4c9a2426336370e6ff78f0d3a19 $
+  $Id: 0e22b0b652743cf4c7a81b992edfa3262ba5dcd0 $
 
   This file, utils.c, contains the implementation of various utility
   functions used in Star Traders.
@@ -53,6 +53,13 @@ wchar_t *mon_thousands_sep;		// Locale's monetary thousands separator
 *                 Module-specific constants and macros                  *
 ************************************************************************/
 
+#define APPIMAGE_NAME		"AppRun.wrapped"
+
+#define DIRSEP			"/"		// Directory separator
+#define CURDIR			"."		// Current directory
+#define HIDDEN_PATH		"."		// Hidden file start char
+#define XDG_DATA_DEFAULT	".local" DIRSEP "share"
+
 #define GAME_FILENAME_PROTO	"game%d"
 #define GAME_FILENAME_BUFSIZE	16
 
@@ -224,6 +231,7 @@ static const unsigned char xor_table[] = {
 
 static char *home_directory_str = NULL;		// Full pathname to home
 static char *data_directory_str = NULL;		// Writable data dir pathname
+static int data_directory_fd    = -1;		// Data dir file descriptor
 
 static bool is_posix_locale = false;		// Override strfmon()?
 
@@ -232,6 +240,31 @@ static bool is_posix_locale = false;		// Override strfmon()?
 *                  Module-specific function prototypes                  *
 ************************************************************************/
 
+/*
+  Function:   home_directory - Return home directory pathname
+  Parameters: (none)
+  Returns:    const char *   - Pointer to home directory
+
+  This function returns the full pathname to the user's home directory,
+  using the HOME environment variable.  Note that the existence of or
+  ability to write to this pathname is NOT checked by this function.
+  NULL is returned if the home directory cannot be determined.
+*/
+static const char *home_directory (void);
+
+
+/*
+  Function:   game_rel_filename - Generate a relative game filename
+  Parameters: gamenum           - Game number (1-9) as an integer
+  Returns:    char *            - Pointer to game filename string
+
+  This function returns a game filename, as a malloc()ed string, relative
+  to some current directory.  If gamenum is outside the range 1 to 9
+  (inclusive), NULL is returned.
+*/
+extern char *game_rel_filename (int gamenum);
+
+
 /*
   Function:   apply_xor - Scramble a buffer using xor_table
   Parameters: dest      - Location of destination buffer
@@ -295,7 +328,8 @@ static ssize_t b64decode (const void *restrict in, size_t inlen,
 *          Initialisation and environment function definitions          *
 ************************************************************************/
 
-// These functions are documented in the file "utils.h"
+/* These functions are documented in the file "utils.h" or in the
+   comments above. */
 
 
 /***********************************************************************/
@@ -306,16 +340,27 @@ void init_program_name (const char *argv0)
     /* This implementation assumes a POSIX environment with an ASCII-safe
        character encoding (such as ASCII or UTF-8). */
 
+    const char *dirsep = DIRSEP;
+
+
+    assert(strlen(dirsep) == 1);
+
     if (argv0 == NULL || *argv0 == '\0') {
 	program_name = PACKAGE;
     } else {
-	char *p = strrchr(argv0, '/');
+	char *p = strrchr(argv0, dirsep[0]);
 
 	if (p != NULL && *++p != '\0') {
 	    program_name = xstrdup(p);
 	} else {
 	    program_name = xstrdup(argv0);
 	}
+
+	// Prevent the AppImage internal name from leaking out
+	if (strcmp(program_name, APPIMAGE_NAME) == 0) {
+	    free((void *) program_name);
+	    program_name = PACKAGE;
+	}
     }
 }
 
@@ -331,6 +376,8 @@ const char *home_directory (void)
 
 	if (home != NULL && *home != '\0') {
 	    home_directory_str = xstrdup(home);
+	} else {
+	    home_directory_str = xstrdup(CURDIR);
 	}
     }
 
@@ -343,28 +390,118 @@ const char *home_directory (void)
 
 const char *data_directory (void)
 {
-    /* This implementation assumes a POSIX environment by using "/" as
+    /* This implementation assumes a POSIX environment by using DIRSEP as
        the directory separator.  It also assumes a dot-starting directory
        name is permissible (again, true on POSIX systems) and that the
        character encoding is ASCII-safe. */
 
+    const char *dirsep            = DIRSEP;
+    const char *dirsep_hiddenpath = DIRSEP HIDDEN_PATH;
+    const char *datahome_part     = DIRSEP XDG_DATA_DEFAULT DIRSEP;
+
+
+    assert(strlen(dirsep) == 1);
+
     if (data_directory_str == NULL) {
 	const char *home = home_directory();
+	const char *xdg_data_home = getenv("XDG_DATA_HOME");
 
-	if (program_name != NULL && home != NULL) {
-	    char *p = xmalloc(strlen(home) + strlen(program_name) + 3);
+	if (program_name != NULL) {
+	    int len_program_name = strlen(program_name);
 
-	    strcpy(p, home);
-	    strcat(p, "/.");
-	    strcat(p, program_name);
-	    data_directory_str = p;
+	    if (home != NULL) {
+		char *p = xmalloc(strlen(home) + strlen(dirsep_hiddenpath)
+				  + len_program_name + 1);
+
+		strcpy(p, home);
+		strcat(p, dirsep_hiddenpath);
+		strcat(p, program_name);
+
+		data_directory_fd = open(p, O_DIRECTORY | O_SEARCH);
+		if (data_directory_fd != -1) {
+		    // Directory "$HOME/.trader" exists: use that
+		    data_directory_str = p;
+		} else {
+		    free(p);
+		}
+	    }
+
+	    if (data_directory_str == NULL && xdg_data_home != NULL
+		&& *xdg_data_home != '\0' && *xdg_data_home == dirsep[0]) {
+		// Use directory "$XDG_DATA_HOME/trader"
+
+		/* Note that the XDG Base Directory Specification v0.7
+		   requires XDG_DATA_HOME to contain an absolute path:
+		   the variable must be ignored if it does not start with
+		   the directory separator DIRSEP ("/") */
+
+		char *p = xmalloc(strlen(xdg_data_home) + strlen(DIRSEP)
+				  + len_program_name + 1);
+
+		strcpy(p, xdg_data_home);
+		strcat(p, dirsep);
+		strcat(p, program_name);
+
+		data_directory_str = p;
+	    }
+
+	    if (data_directory_str == NULL && home != NULL) {
+		// Use directory "$HOME/.local/share/trader"
+
+		char *p = xmalloc(strlen(home) + strlen(datahome_part)
+				  + len_program_name + 1);
+
+		strcpy(p, home);
+		strcat(p, datahome_part);
+		strcat(p, program_name);
+
+		data_directory_str = p;
+	    }
 	}
     }
 
+    if (data_directory_str != NULL && data_directory_fd == -1) {
+	data_directory_fd = open(data_directory_str, O_DIRECTORY | O_SEARCH);
+    }
+
     return data_directory_str;
 }
 
 
+/***********************************************************************/
+// data_directory_fileno: Return file descriptor of data directory
+
+int data_directory_fileno (void)
+{
+    if (data_directory_fd == -1) {
+	(void) data_directory();
+    }
+
+    return data_directory_fd;
+}
+
+
+/***********************************************************************/
+// game_rel_filename: Generate a relative game filename
+
+char *game_rel_filename (int gamenum)
+{
+    /* This implementation assumes a POSIX environment and an ASCII-safe
+       character encoding. */
+
+    char *p;
+
+
+    if (gamenum < 1 || gamenum > 9) {
+	return NULL;
+    }
+
+    p = xmalloc(GAME_FILENAME_BUFSIZE);
+    snprintf(p, GAME_FILENAME_BUFSIZE, GAME_FILENAME_PROTO, gamenum);
+    return p;
+}
+
+
 /***********************************************************************/
 // game_filename: Convert an integer to a game filename
 
@@ -373,30 +510,72 @@ char *game_filename (int gamenum)
     /* This implementation assumes a POSIX environment and an ASCII-safe
        character encoding. */
 
-    char buf[GAME_FILENAME_BUFSIZE];	// Buffer for part of filename
+    const char *dirsep = DIRSEP;
+
+    char *gfn;				// Relative game filename
     const char *dd;			// Data directory
 
-
-    if (gamenum < 1 || gamenum > 9) {
+    gfn = game_rel_filename(gamenum);
+    if (gfn == NULL) {
 	return NULL;
     }
 
     dd = data_directory();
-    snprintf(buf, GAME_FILENAME_BUFSIZE, GAME_FILENAME_PROTO, gamenum);
-
     if (dd == NULL) {
-	return xstrdup(buf);
+	return gfn;
     } else {
-	char *p = xmalloc(strlen(dd) + strlen(buf) + 2);
+	char *p = xmalloc(strlen(dd) + strlen(dirsep) + strlen(gfn) + 1);
 
 	strcpy(p, dd);
-	strcat(p, "/");
-	strcat(p, buf);
+	strcat(p, dirsep);
+	strcat(p, gfn);
+	free(gfn);
 	return p;
     }
 }
 
 
+/***********************************************************************/
+// game_fopen: Open a game file for reading or writing
+
+FILE *game_fopen (int gamenum, const char *mode)
+{
+    char *gfn;
+    int ddfd;
+    int fd;
+    int oflag;
+    mode_t omode;
+
+
+    gfn = game_rel_filename(gamenum);
+    if (gfn == NULL) {
+	return NULL;
+    }
+
+    ddfd = data_directory_fileno();
+    if (ddfd == -1) {
+	ddfd = AT_FDCWD;
+    }
+
+    if (strcmp(mode, "r") == 0) {
+	oflag = O_RDONLY;
+	omode = 0;
+    } else if (strcmp(mode, "w") == 0) {
+	oflag = O_WRONLY | O_CREAT | O_TRUNC;
+	omode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
+    } else {
+	free(gfn);
+	errno = EINVAL;
+	return NULL;
+    }
+
+    fd = openat(ddfd, gfn, oflag, omode);
+    free(gfn);
+
+    return (fd == -1) ? NULL : fdopen(fd, mode);
+}
+
+
 /************************************************************************
 *                 Error-reporting function definitions                  *
 ************************************************************************/
@@ -510,9 +689,31 @@ extern int randi (int limit)
 
 
 /***********************************************************************/
-// init_locale: Initialise locale-specific variables
+// init_locale: Initialise locale and message catalogs
 
-void init_locale (void)
+extern void init_locale (void)
+{
+    const char *podir = getenv("TEXTDOMAINDIR");
+
+
+    // Initialise the current locale
+    if (setlocale(LC_ALL, "") == NULL) {
+	err_exit("could not set locale "
+		 "(check LANG, LC_ALL and LANGUAGE in environment)");
+    }
+
+    // Use correct message catalogs for the locale
+    bindtextdomain(
+	PACKAGE,
+	(podir != NULL && *podir != '\0') ? podir : LOCALEDIR);
+    textdomain(PACKAGE);
+}
+
+
+/***********************************************************************/
+// init_locale_vars: Initialise locale-specific variables
+
+void init_locale_vars (void)
 {
     char *cur, *cloc;
     struct lconv *lc;
@@ -728,6 +929,7 @@ char *scramble (char *restrict dest, const char *restrict src,
 	// Place checksum in front of Base64 string
 	snprintf(chksumbuf, SCRAMBLE_CHKSUM_LEN + 1, "%03x", chksum);
 	memcpy(dest, chksumbuf, SCRAMBLE_CHKSUM_LEN);
+	dest[size - 1] = '\0';		// Keep Coverity Scan happy...
 
 	free(xorbuf);
     }
@@ -903,7 +1105,7 @@ size_t b64encode (const void *restrict in, size_t inlen,
 
     if (padding > 0) {
 	assert(count + 2 < outlen);
-	for (; padding < 3; padding++) {
+	for ( ; padding < 3; padding++) {
 	    *u_out++ = SCRAMBLE_PAD_CHAR;
 	    count++;
 	}
@@ -1003,6 +1205,69 @@ ssize_t b64decode (const void *restrict in, size_t inlen,
 // These functions are documented in the file "utils.h"
 
 
+/***********************************************************************/
+// xmkdir: Check and create directory with its parents
+
+int xmkdir (const char *pathname, mode_t mode)
+{
+    const char *dirsep = DIRSEP;
+    struct stat statbuf;
+    char *pathcopy;
+    char *pcomp, *pend;
+    int ret;
+
+
+    assert(strlen(dirsep) == 1);
+
+    if (pathname == NULL || *pathname == '\0') {
+	errno = ENOENT;				// As documented by POSIX
+	return -1;
+    }
+
+    // Try creating the directory
+    ret = mkdir(pathname, mode);
+    if (ret != 0 && errno == EEXIST) {
+	// Pathname exists: is it a directory?
+	if (stat(pathname, &statbuf) == 0) {
+	    if (S_ISDIR(statbuf.st_mode)) {
+		return 0;
+	    } else {
+		errno = ENOTDIR;
+		return -1;
+	    }
+	} else {
+	    return -1;
+	}
+    } else if (ret == 0 || (errno != ENOENT && errno != ENOTDIR)) {
+	return ret;
+    }
+
+    // Try creating directory components, except the last, one by one
+    pathcopy = xstrdup(pathname);
+    pcomp = pend = pathcopy;
+    for ( ; *pend != '\0'; pend++) {
+	if (*pend == dirsep[0] && pcomp != pend) {
+	    *pend = '\0';
+
+	    ret = mkdir(pathcopy, mode);
+	    if (ret != 0 && errno != EEXIST) {
+		free(pathcopy);
+		return ret;
+	    }
+
+	    *pend = dirsep[0];
+	    pcomp = pend + 1;
+	}
+    }
+
+    // Try creating the last directory component
+    ret = mkdir(pathcopy, mode);
+
+    free(pathcopy);
+    return ret;
+}
+
+
 /***********************************************************************/
 // xmalloc: Allocate a new block of memory, with checking
 
